# dclock: program to display time on a 4 x 8x8 LED Matrix with max7219 drivers
# Using ESP32 NodeMCU board with micropython v1.17
# ESP32 RTC time is synced with time of ntp server twice a day
# time(s) of sync can be set in variable STATE['sync_at'], multiple times can be set in a list
# sycing time is in UTC time(s)
#
# module utime is part of micropython
# modules wifiConnect, ntptime, max7219 and this file, dclock.py, need to be downloaded to the ESP32 so they can be imported
#
# Set your wlan credetials, SSID and password, in module wlanConnect.py
#
# Multiple lines are commented out in the functions and main loop, 
# remove # as needed for debugging
#
# max7219.py is Copyright (c) 2017 Mike Causer under MIT license. See github page: https://github.com/mcauser/micropython-max7219 for current license information and details

from machine import Pin, SPI, ADC, RTC
import max7219
import wlanConnect
import utime
import ntptime
import time

# Daylight Savings switch is on Pin 13 (D7), internal pull up enabled
DSpin = Pin(23, Pin.IN, Pin.PULL_UP)
#
# instantiate ADC to read LDR voltage on GPIO34 with 11dB attenuation for 3.6V range
light_value = ADC(Pin(34))
light_value.atten(ADC.ATTN_11DB)
#
rtc = RTC()
spi = SPI(1, baudrate=10000000, polarity=1, phase=0, sck=Pin(14), mosi=Pin(13))
cselect = Pin(27, Pin.OUT)
display = max7219.Matrix8x8(spi, cselect, 4)

# dictionary STATE holds and initilizes main variables
STATE = {
    'hour': 0,
    'minute': 0,
    'UTCdiff' : +1,
    'time_checked' : 0,
    'utc_hour' : 0,
    'utc_minute' : 1,
    'sync_at' : [8,20]
    }

tick=0  # initialize life tick counter

# Function set_brightness() calculates the intensity value for the LEDs based on ADC reading (0...4096)
# Reading of 4069 refers to LDR is in totally dark environment
# Any non-linearity of LDR or lighting in the room can be adjusted in below "if-elif-chain"
# Display brightness range between 0 and 15, 
# Test showed brightness 3 is enough for normal room lighting condition, adjust for your room
#
# I am using only four steps to set brightness for dark - low - medium - bright lighting
# set as many as you need by adding elif blocks
# uncomment line #print('ADCreading= ',dim) to display ADC to see your lighting conditions
def set_brightness():
    
    dim = light_value.read()
    #print('ADCreading= ',dim)

    if dim > 3000:  #dark
        display.brightness(0)

    elif dim > 1500: #low
        display.brightness(1)

    elif dim > 500:  #medium
        display.brightness(2)

    else:   #bright
        display.brightness(3)
#

# display functions
def clear_display():
    display.fill(0)
    display.show()

def display_on():
    display.brightness(5)
    display.fill(1)
    display.show()

# life tick on display, two versions
def life_tick1():
    global tick
    if not tick:
        display.pixel(16,3,1)
        display.pixel(15,4,1)
        display.pixel(16,4,0)
        display.pixel(15,3,0)
        display.show()
        tick=1
    else:
        display.pixel(16,3,0)
        display.pixel(15,4,0)
        display.pixel(16,4,1)
        display.pixel(15,3,1)
        display.show()
        tick=0

def life_tick():
    global tick
    if not tick:
        display.pixel(16,3,1)
        display.pixel(16,4,1)
        display.show()
        tick=1
    else:
        display.pixel(16,3,0)
        display.pixel(16,4,0)

        display.show()
        tick=0


# time related functions
def sync_time():

    #read time from ntp server and set RTC to curtrent time
    #connect and disconnect from wifi here to avoid stoping word clock on wifi errors, which can occur while wifi is always connected
    #
    #utime returns 8 tuple YYYY, M, D, H, Min, Sec, e.g. (2018, 2, 3, 18, 3, 2, 5, 34)

    # uncomment below line for debugging
    #print("syncing time")
    
    try:

        wlanConnect.connect()
        #
        currenttime = utime.localtime(ntptime.time())
        print('UTC is ',currenttime)

        ntpyear = currenttime[0]
        ntpmonth = currenttime[1]
        ntpday = currenttime[2]
        ntphour = currenttime[3]
        ntpminute = currenttime[4]
        ntpsecond = currenttime[5]

        # set ESP32 RTC to ntp time = UTC
        # weekday and millisecond set to 0
        rtc.datetime((ntpyear, ntpmonth, ntpday, 0, ntphour, ntpminute, ntpsecond, 0))

    except OSError as oserr:
        # catch timeout acessing the internet
        print("OSError exception: ", oserr)
        time.sleep(3)

    except:
        # catch any other exceptions
        print("other exception: ")
        time.sleep(1)
        reset() #restart ESP32 in case of e.g. wlan connection does not work

    wlanConnect.disconnect()


def update_time():

    #print("update_time")
    # read time from RTC
    # we get (year, month, day, weekday, hour, minute, second, millisecond)
    datetime = rtc.datetime()
    print('ESP32 RTC: ',datetime)

    STATE['utc_hour'] = datetime[4]
    STATE['utc_minute'] = datetime[5]

    # uncomment two lines below for manual time entry while debugging
    #STATE['utc_hour'] = int(input("hour?: "))
    #STATE['utc_minute'] = int(input("minute?: "))  

    # check for Daylight savings switch (DSpin)
    # we get UTC from ntp, so need to correct for Time zone
    # time zone correction as set in STATE['UTCdiff']
    
    if DSpin.value() == 1: #daylight savings on

        STATE['hour'] = (STATE['utc_hour'] + STATE['UTCdiff'] + 1) % 24
    else:
        STATE['hour'] = (STATE['utc_hour'] + STATE['UTCdiff']) % 24

        
    STATE['minute'] = STATE['utc_minute']


    # at times defined in STATW['sync_at'] sync RTC with ntp time
    # tracking if time is already checked avoids syncing many times in that hour
    if (STATE['hour'] in STATE['sync_at']):
        if STATE['time_checked'] == 0:
            sync_time()
            STATE['time_checked'] = 1
    
    # reset time_checked outside sync hours
    if STATE['time_checked'] ==1:
        if not (STATE['hour'] in STATE['sync_at']):
            STATE['time_checked'] = 0

    #print('time checked; ', STATE['time_checked'])
#

# start main program

# first connection to wlan network
display.text('wlan',0,0,1)
display.show()
wlanConnect.connect()

# first time syncing RTC with internet time
clear_display()
display.text('sync',0,0,1)
display.show()
sync_time()
time.sleep_ms(500)


#main loop
while True:
    update_time()
    #
    print("Display time {}:{}".format(STATE['hour'], STATE['minute']))
    display.fill(0)    #clear display buffer

    # format time for LED matrix
    # add a leading "0" to minute if minute is one digit only (<10)
    if STATE['minute'] < 10:
        now = "{0:2d}0{1:1d}".format(STATE['hour'], STATE['minute'])
    else:
        now = "{0:2d}{1:2d}".format(STATE['hour'], STATE['minute'])

    #print('now= ', now)
    display.text(now,0,0,1)
    display.show()
    
    # read LDR and adjust intensity of LED matrix display
    set_brightness()

    # show life tick on display, two versions life_tick() or life_tick1()
    life_tick()
    
    # wait a sec
    time.sleep_ms(1000)

    # end main loop


# end program
